/*

 Copyright (c) 2001 The Regents of the University of California.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

 */

package flathtlc;

import flathtlc.SymbolTable;
import flathtlc.analysis.DepthFirstAdapter;
import flathtlc.node.AModeDeclaration;
import flathtlc.node.Token;

/**
 * 
 * This class checks whether the input program is 'time-safe'.
 * 
 * @author Christoph Kirsch, cm@eecs.berkeley.edu
 */
public class TimeSafetyChecker extends DepthFirstAdapter {
//	private int modePeriod = 0;

	private double modeUtilization = 0.0;

	private boolean noWcet = true;

	private SymbolTable symbolTable;

	public TimeSafetyChecker(SymbolTable symbolTable) {
		this.symbolTable = symbolTable;
	}

	public void inAModeDeclaration(AModeDeclaration node) {

		modeUtilization = 0.0;
	}

	public void outAModeDeclaration(AModeDeclaration node) {
		if (modeUtilization > 1.0)
			errorNotTimeSafe(node.getModeName());
		else if (!noWcet)
			System.out.println("Mode " + node.getModeName().getText()
					+ ", utilization: " + modeUtilization);

		modeUtilization = 0.0;
	}

	private void errorNotTimeSafe(Token modeToken) {
		throw new RuntimeException("[" + modeToken.getLine() + ","
				+ modeToken.getPos() + "]" + " Mode " + modeToken.getText()
				+ " not time safe, utilization: " + modeUtilization);
	}
	
	public String toString() {
		return symbolTable.toString();
	}
}
